# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fasttext
import langid
import cld3
import os
import urllib.request
import logging


class LangDetect:
  FAST_TEXT_MODEL_URL = 'https://storage.googleapis.com/babel-tower-exp/model/fasttext/langdetect/lid.176.ftz'
  
  def __init__(self, model_path = '/tmp/lid.176.ftz'):
    if not os.path.exists(model_path):
      print(f'downloading model from {self.FAST_TEXT_MODEL_URL} => {model_path}')
      with urllib.request.urlopen(self.FAST_TEXT_MODEL_URL) as f:
        with open(model_path, 'wb') as output:
          output.write(f.read())      
    self.ft_model = fasttext.load_model(model_path)
  
  def fasttext_get_language(self, text):
    labels, scores = self.ft_model.predict(text, k=5)
    return [(label.replace('__label__', ''), score) for (label, score) in zip(labels, scores)]

  def langid_get_language(self, text):
    lang, score = langid.classify(text)
    return [(lang, 1.0)]

  def cld3_get_language(self, text):
    result = cld3.get_language(text)
    if result:
      return [(result.language, result.probability)]
    else:
      logging.warn('cld3 failed to get predictions for text: %s', text)
      return []

  def get_language(self, text, return_details=False):
    if '\n' in text:
      text = text.split('\n')[0]
    
    detail_results = [
      self.langid_get_language(text),
      self.fasttext_get_language(text),
      self.cld3_get_language(text)
    ]
    scores = {}
    for result in detail_results:
      for lang, score in result:
        scores[lang] = scores.get(lang, 0) + score 

    result = None
    max_score = 0
    for lang, score in scores.items():
      if score > max_score:
        result = lang
        max_score = score
    
    if return_details:
      return result, detail_results
    else:
      return result


# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from transformers import AutoTokenizer, AutoModelForSeq2SeqLM

from langdetect import detect

import six
from google.cloud import translate_v2 as translate


# TODO: support cloud translation API


# From langdetect language to huggingface model language
# ref: https://pypi.org/project/langdetect/
# ref: https://huggingface.co/models?filter=translation
LANG_MAPPING = {
  'zh-cn': 'zh',
  'zh-tw': 'zh',
  'jp': 'jap',
}


class TranslationClient:
  def __init__(self, engine='hf'):
    """Initializes translation client.
    
    Args: 
      engine: translation engine. valid options are:
        'hf' - using hugging face pipeline
        'gcp' - using Google Cloud translation API
    """
    self.engine = engine
    self.model_cache = {}
    
  def _get_hf_model(self, src_lang, target_lang):
    model_id = f"Helsinki-NLP/opus-mt-{src_lang}-{target_lang}"
    if model_id not in self.model_cache:
      print('loading model:', model_id)
      tokenizer = AutoTokenizer.from_pretrained(model_id)
      model = AutoModelForSeq2SeqLM.from_pretrained(model_id)
      self.model_cache[model_id] = (tokenizer, model)
    
    return self.model_cache[model_id]
    
  def _translate_hf(self, sentence, src_lang, target_lang):
    tokenizer, model = self._get_hf_model(src_lang, target_lang)
    inputs = tokenizer.encode(sentence, return_tensors="pt")
    outputs = model.generate(inputs, max_length=40, num_beams=4, early_stopping=True)
    return tokenizer.decode(outputs[0]).replace('<pad>', '')

  def translate(self, sentence, target_lang, src_lang=None):
    """Translates a sentence.
    
    Args:
      sentence: The sentence to translate.
      target_lang: translate to this language.
      src_lang: translate from this language. If none, automatically determine source language.
    """
    if self.engine == 'hf':
      if src_lang is None:
        src_lang = detect(sentence)
      
      src_lang = LANG_MAPPING.get(src_lang, src_lang)
      target_lang = LANG_MAPPING.get(target_lang, target_lang)
      
      if src_lang == target_lang:
        # skip translation if src and target are the same
        return sentence
      
      try:
        result = self._translate_hf(sentence, src_lang, target_lang)
      except:
        en_sentence = self._translate_hf(sentence, src_lang, 'en')
        print(src_lang, '->', 'en:', sentence, en_sentence)
        result = self._translate_hf(en_sentence, 'en', target_lang)
        print('en', '->', target_lang, en_sentence, result)
      return result
    elif self.engine == 'gcp':
      translate_client = translate.Client()
      if isinstance(sentence, six.binary_type):
        sentence = sentence.decode("utf-8")
      # Text can also be a sequence of strings, in which case this method
      # will return a sequence of results for each text.
      try:
        result = translate_client.translate(sentence, target_language=target_lang)
        return result["translatedText"]
      except:
        print('Unsupported language', target_lang)
        return
    else:
      raise RuntimeError('Unsupported')


# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import spacy_keyword_extraction
from rake_nltk import Rake
from keybert import KeyBERT
from gps_babel_tower.tasks.lang_detect import LangDetect
from gps_babel_tower.tasks.word_seg import WordSegment
import logging
import rake_ja


class KeywordExtraction:
  def __init__(self, kw_model, min_len=1, max_len=3):
    self.kw_model = kw_model
    self.min_len = min_len
    self.max_len = max_len
    self.langdetect = LangDetect()
    self.word_seg = WordSegment()
    
    self.ja_tok = rake_ja.Tokenizer()
    self.ja_rake = rake_ja.JapaneseRake()
    
    if kw_model == 'rake':
      self.model = Rake(min_length=min_len, max_length=max_len)
    elif kw_model == 'keybert':
      self.model = KeyBERT('xlm-r-100langs-bert-base-nli-mean-tokens')
    elif kw_model == 'spacy_bigram':
      self.model = spacy_keyword_extraction.KeywordExtraction()
    else:
      raise Exception(f'Unsupported keyword extraction model type: {kw_model}, can only be rake|keybert|spacy_bigram')
      
  
  def extract_keywords(self, text, lang=None, max_results=-1):
    if not lang:
      lang = self.langdetect.get_language(text)

    if self.kw_model == 'rake':
      if lang == 'ja':
        tokens = self.ja_tok.tokenize(text)
        self.ja_rake.extract_keywords_from_text(tokens)
        keywords = self.ja_rake.get_ranked_phrases()
      else:
        self.model.extract_keywords_from_text(text)
        keywords = self.model.get_ranked_phrases()
    elif self.kw_model == 'keybert':
      if lang == 'ja' or lang == 'zh':
        text = self.word_seg.tokenize(text, lang=lang, join_by=' ')
      keywords_and_score = self.model.extract_keywords(text, keyphrase_ngram_range=(self.min_len, self.max_len))
      keywords = [keyword for keyword, _score in keywords_and_score]
    elif self.kw_model == 'spacy_bigram':
      if lang != 'en':
        logging.error(f'spacy_bigram only supports English! language detected: {lang} for text {text}')
        return []
      keywords = self.model.tokenize(text)
      
    return keywords[:max_results]


# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import os
import logging
from datasets import load_dataset
from transformers import AutoTokenizer, DPRContextEncoder


_EMB_COL = 'embedding'
_TEXT_COL = 'text'


class NearestNeighbor:
  def __init__(self,
               model_id,
               data_file_path,
               force_rebuild_index=False,
               index_file_path=None,
              ):
    self.tokenizer = AutoTokenizer.from_pretrained(model_id)
    self.model = DPRContextEncoder.from_pretrained(model_id)
    
    ds = load_dataset('text', data_files={'train': data_file_path})['train']
    if (not force_rebuild_index) and index_file_path and os.path.exists(index_file_path):
      logging.info('Load index file from %s', index_file_path)
      ds.load_faiss_index('embedding', index_file_path)
    else:
      logging.info('Building index')
      ds = ds.map(lambda example: {_EMB_COL: self.encode(example[_TEXT_COL])})
      ds.add_faiss_index(column=_EMB_COL)
      if index_file_path:
        logging.info('Save index to %s', index_file_path)
        ds.save_faiss_index(_EMB_COL, index_file_path)
        
    self.dataset = ds

  def encode(self, text):
    tokens = self.tokenizer(text, return_tensors="pt")
    with torch.no_grad():
      result = self.model(**tokens).pooler_output.numpy()[0]
    return result
  
  def get_nearest_neighbors(self, query, top_k=10):
    q_emb = self.encode(query)
    scores, retrieved_examples = self.dataset.get_nearest_examples(_EMB_COL, q_emb, k=top_k)
    return retrieved_examples[_TEXT_COL]



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from google.cloud import translate_v2 as translate


class LanguageDetection():
  def __init__(self, engine='gcp'):
    self.engine = engine
    if self.engine == 'gcp':
      self.translate_client = translate.Client()
    else:
      raise RuntimeError(f'engine {engine} not supported yet')
    return

  def detect(self, text):
    language_code = None
    if self.engine == 'gcp':
      result = self.translate_client.detect_language(text)
      # print("Confidence: {}".format(result["confidence"]))
      language_code = result["language"]
    else:
      raise RuntimeError('Unsupported')
    return language_code




# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from transformers import AutoTokenizer, AutoModelForSequenceClassification
import torch
from google.cloud import language_v1 as language
import logging

# Explore models here:
# https://huggingface.co/models?search=sentiment


class SentimentClient:
  def __init__(self, engine='hf', use_gpu=False, model='distilbert-base-uncased-finetuned-sst-2-english', use_fast=True):
    self.engine = engine
    if engine == 'hf':
      self.tokenizer = AutoTokenizer.from_pretrained(model, use_fast=use_fast)
      self.model = AutoModelForSequenceClassification.from_pretrained(model)
      if use_gpu:
        self.model.cuda()
      self.use_gpu = use_gpu
    elif engine == 'gcp':
      self.client = language.LanguageServiceClient()
      
    else:
      raise RuntimeError(f'engine {engine} not supported yet')

  def score(self, sentence, score_range=(-1,1)):
    if self.engine == 'hf':
      tokens = self.tokenizer(sentence, padding=True, truncation=True, max_length=128, return_tensors="pt")
      if self.use_gpu:
        tokens = {k: v.cuda() for k,v in tokens.items()}
      with torch.no_grad():
        logits = self.model(**tokens).logits
      probs = torch.softmax(logits, dim=-1)
      max_score = probs.shape[1]-1
      scores_arange = torch.arange(max_score+1)
      if self.use_gpu:
        scores_arange = scores_arange.cuda()
      score = torch.sum(scores_arange * probs, axis=-1) # (0, max_score)
      score = score_range[0] + (score_range[1]-score_range[0]) * score / max_score
      if self.use_gpu:
        score = score.cpu()
      score = score.numpy()
      if score.size == 1:
        score = score.item()
      return score
    elif self.engine == 'gcp':
      document = {
        'content':sentence,
        "type_":language.Document.Type.PLAIN_TEXT
        }
      encoding_type = language.EncodingType.UTF8
      
      request = {'document': document, 'encoding_type': encoding_type}
      try:
        # Detects the sentiment of the text
        response = self.client.analyze_sentiment(request = request)
      except:
        logging.exception('unable to get score from gcp')
        return 0
      score = response.document_sentiment.score
      # score origin range(-1,1)
      score = score_range[0] + (score_range[1]-score_range[0]) * ((score + 1) / 2)
      return score
    else:
      raise RuntimeError(f'Unsupported engine {self.engine}')


# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import shutil

from setuptools import find_packages, setup
from setuptools.command.install import install
from subprocess import getoutput


install_requires = [
  'nltk~=3.2.5',
  'transformers~=4.6.1',
  'tensorflow~=2.4.1',
  'torch~=1.8.1',
  'langdetect>=1.0.8',
  'sentencepiece>=0.1.85',
  'datasets~=1.5.0',
  'faiss-cpu~=1.7.0',
  'google-cloud-translate>=3.0.2',
  'google-cloud-language>=2.0.0',
  'fasttext>=0.9.2',
  'langid>=1.1.6',
  'pycld3>=0.20',
  'gspread>=3.6.0',
  'gspread-dataframe>=3.2.0',
  'spacy~=3.0.5',
  'rake-nltk~=1.0.4',
  'keybert~=0.3.0',
  'jieba~=0.42.1',
  'nagisa~=0.2.7',
  'rake-ja~=0.0.1',
]


setup(
  name="gps-babel-tower",
  version="0.0.2", # expected format is one of x.y.z.dev0, or x.y.z.rc1 or x.y.z (no to dashes, yes to dots)
  author="Chi Zhang",
  author_email="chii@google.com",
  description="Babel Tower NLP Library",
  long_description=open("README.md", "r", encoding="utf-8").read(),
  long_description_content_type="text/markdown",
  license="Apache",
  package_dir={"": "src"},
  packages=find_packages("src"),
  extras_require={},
  python_requires=">=3.6.0",
  install_requires=install_requires
)


# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jieba
import nagisa

class WordSegment:
  def tokenize(self, text, lang, join_by=None):
    if lang == 'zh':
      words = jieba.cut(text)
    elif lang == 'ja':
      words = nagisa.tagging(text)
      words = words.words
    else:
      raise Exception(f'Unknown language {lang}')
    
    if join_by:
      return join_by.join(words)
    


# coding=utf-8
# Copyright 2020 Google LLC..
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Keywords extraction component."""
import re
import logging
import string

import nltk
from nltk.corpus import stopwords
import spacy
import spacy.cli.download


class KeywordExtraction:

  def __init__(self):
    try:
      self.nlp = spacy.load('en_core_web_sm')
    except:
      logging.exception('load model failed, download automatically') 
      spacy.cli.download('en_core_web_sm')
      self.nlp = spacy.load('en_core_web_sm')
    
    nltk.download('stopwords')
    self.stopwords_en = set(stopwords.words('english'))

  def tokenize(self, text, stop_words=None, word_normalize_map=None, mode=None):
    sentence_token_limit = 1
    lda_tokens = []
    tokens = self.nlp(self.format_sentence(text))

    if stop_words:
      self.stopwords_en.update(stop_words)

    if len(tokens) >= sentence_token_limit:
      root_word = ''
      for token in tokens:
        if token.orth_.isspace():
          continue
        if token.lower_ in self.stopwords_en:
          continue
        words = None
        if mode == 'entity_vector':
          if token.dep_ in ('nsubj', 'nsubjpass'):
            words = token.vector
        elif mode == 'entity_name':
          if token.dep_ in ('nsubj', 'nsubjpass'):
            words = token.lower_
        else:
          if token.dep_ in ('compound', 'dobj', 'pobj', 'xcomp', 'nsubj',
                            'nsubjpass', 'advcl', 'amod', 'acomp'):
            word = self.normalize_word(word_normalize_map, token)
            if token.dep_ in ('dobj', 'pobj', 'xcomp'):
              if token.dep_ in ('pobj',):
                words = self.normalize_word(word_normalize_map,
                                            token.head.head) + ' ' + word
              else:
                words = self.normalize_word(word_normalize_map,
                                            token.head) + ' ' + word
            elif token.dep_ in ('compound', 'advcl', 'nsubj', 'nsubjpass',
                                'amod', 'acomp'):
              if token.head.lemma_ == 'be':
                if token.dep_ in ('nsubj', 'nsubjpass'):
                  root_word = word
                  root_word_head_idx = token.head.idx
                elif (root_word and
                      root_word_head_idx == token.head.idx and
                      token.dep_ == 'acomp'):
                  words = word + ' ' + root_word
              elif token.dep_ in ('compound', 'advcl', 'nsubj', 'nsubjpass',
                                  'amod'):
                words = word + ' ' + self.normalize_word(
                    word_normalize_map,
                    token.head)
        if words is not None:
          lda_tokens.append(words)
    return lda_tokens

  def format_sentence(self, text, stop_words=None, strip_all=False):
    if isinstance(text, str):
      emoji_pattern = re.compile(
          '['
          u'\U0001F600-\U0001F64F'  # emoticons
          u'\U0001F300-\U0001F5FF'  # symbols & pictographs
          u'\U0001F680-\U0001F6FF'  # transport & map symbols
          u'\U0001F1E0-\U0001F1FF'  # flags (iOS)
          ']+',
          flags=re.UNICODE)
      text = emoji_pattern.sub(r'', text)
      if stop_words:
        text = ' '.join([
            word for word in nltk.word_tokenize(text)
            if word.lower() not in stop_words
        ])
      if strip_all:
        del_str = string.punctuation + string.digits  # ASCII 标点符号
      else:
        del_str = string.digits
      identify = str.maketrans('', '', del_str)
      clean_line = text.translate(identify)
      return clean_line
    else:
      print('text format error:', text)
      return ''

  def normalize_word(self, words_map, token):
    if words_map is not None and token.text in words_map:
      normalized_word = words_map[token.text].lower()
    else:
      normalized_word = token.lemma_
    return normalized_word




# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
# also try: dpr-question_encoder-multiset-base
from transformers import AutoTokenizer, DPRQuestionEncoder, TFDPRQuestionEncoder
from transformers import XLMTokenizer, TFXLMModel
import tensorflow as tf
from typing import Sequence, Union
import enum


class EncoderType(enum.Enum):
  DPR = enum.auto()
  XLM = enum.auto()


class TextEncoder:
  # Ref: https://huggingface.co/transformers/multilingual.html
  def __init__(self, model=EncoderType.XLM, max_length=32):
    if model == EncoderType.DPR:
      model_key = 'facebook/dpr-question_encoder-single-nq-base'
      self.tokenizer = AutoTokenizer.from_pretrained(model_key)
      self.model = TFDPRQuestionEncoder.from_pretrained(model_key)
    elif model == EncoderType.XLM:
      model_key = 'xlm-mlm-100-1280'
      self.tokenizer = XLMTokenizer.from_pretrained(model_key)
      self.model = TFXLMModel.from_pretrained(model_key)
    print('using model', model_key)
    self.max_length = max_length
    self.emb_cache = {}

  def encode(self, words, cache_key=None, max_length=None):
    if cache_key and cache_key in self.emb_cache:
      return self.emb_cache[cache_key]
    
    max_length = max_length or self.max_length
    tokens = self.tokenizer(words, padding='max_length', max_length=max_length, truncation=True, return_tensors='tf')
    outputs = self.model(**tokens)
    if hasattr(outputs, 'pooler_output'):
      result = outputs.pooler_output
    else:
      # use first token output as sentence encoding
      # TODO: try average pooling
      result = outputs.last_hidden_state[:,0,:]
      
    if cache_key:
      self.emb_cache[cache_key] = result
    
    return result
  
  def batch_similarity(self,
                       query: Union[str, Sequence[str]],
                       candidates: Sequence[str],
                       cache_key1=None,
                       cache_key2=None,
                       similarity='cosine',
                       return_type='list'):
    """Compute 1 vs n similarity from query to a list of candidates."""
    return_single = False
    if isinstance(query, str):
      query = [query]
      return_single = True
    
    # (query_count, dim)
    query_embedding = self.encode(query, cache_key=cache_key1)
    # (candidates_count, dim)
    candidates_embedding = self.encode(candidates, cache_key=cache_key2)
    
    if similarity == 'cosine':
      sim_func = cosine_similarity
    elif similarity == 'l2':
      sim_func = l2_distance
    else:
      raise Exception(f'unknown similarity function {similarity}')
      
    # (candidates_count,)
    result = sim_func(query_embedding, candidates_embedding)
    
    if return_type == 'raw':
      pass
    elif return_type == 'list':
      result = list(result.numpy())
    elif return_type == 'numpy':
      result = result.numpy()
    else:
      raise Exception(f'unknown return type {return_type}')
      
    if return_single:
      result = result[0]
      
    return result

  
def cosine_similarity(v1, v2, eps=1e-8):
  """v1:  (n1, k)  vector
     v2:  (n2, k)  vector
     Returns:  (n1, n2)  cosine similarities
  """
  p = tf.matmul(v1, v2, transpose_b=True)
  n = tf.math.maximum(tf.expand_dims(tf.norm(v1, axis=-1), axis=-1) * tf.expand_dims(tf.norm(v2, axis=-1), axis=0), eps)
  scores = p / n
  return scores


def l2_distance(v1, v2):
  """v1:  (n1, k)  vector
     v2:  (n2, k)  vector
     Returns:  (n1, n2)  l2 distances
  """
  return tf.norm(tf.expand_dims(v1, 1) - tf.expand_dims(v2, 0), axis=-1)
  
